from tempfile import NamedTemporaryFile, TemporaryDirectory
import shutil
import subprocess
import sys
import os
from itertools import chain

import numpy as np
from forwardable import forwardable
import matplotlib.pyplot as plt
from PIL import Image
import json
import msgpack

import homcloud.alpha_filtration as alpha_filtration
import homcloud.pict.binarize as binarize
import homcloud.diagram as homdiag
import homcloud.plot_PD as plot_PD
import homcloud.bitmap_complex as bitmap_complex
import homcloud.full_ph_tree as full_ph_tree
import homcloud.index_map_advanced as index_map
import homcloud.utils as utils
from homcloud.optimal_cycle import VolumeOptimalCycleFinder
from homcloud.full_ph_tree import SpatialSearcher
import homcloud.vectorize_PD as vectorize_PD
import homcloud.view_index_pict as view_index_pict
import homcloud.plot_PD_slice as plot_PD_slice
import homcloud.pict.tree as pict_tree


_tempdir_path = None


def tempdir():
    global _tempdir_path
    if not _tempdir_path:
        _tempdir_path = TemporaryDirectory()

    return _tempdir_path


def tempfile(suffix):
    file = NamedTemporaryFile(suffix=suffix, dir=tempdir().name, delete=False)
    file.close()
    return file.name


def example_data(name):
    if name == "tetrahedron":
        return np.array([
            [0.0, 0.0, 0.0],
            [8.0, 0.0, 0.0],
            [5.0, 6.0, 0.0],
            [4.0, 2.0, 6.0],
        ])
    raise ValueError("Unknown example name: %s" % name)


class PDs(object):
    def __init__(self, path, type=None):
        self.path = path
        self.type = type
        self.diagrams = dict()
        self.index_map = None

    def __repr__(self):
        return "PDs(path=%s)" % self.path

    @staticmethod
    def from_alpha_filtration(pointcloud,
                              weight=False, no_squared=False,
                              subsets=False, check_acyclicity=False,
                              parallels=1, save_to=None):
        input_data = pointcloud.astype(dtype=float)
        num_points, dim = pointcloud.shape
        if isinstance(weight, np.ndarray):
            input_data = np.hstack([input_data,
                                    weight.reshape(num_points, 1)])
            weight = True
        elif weight is True:
            dim -= 1

        if isinstance(subsets, np.ndarray):
            input_data = np.hstack([input_data,
                                    subsets.reshape(num_points, 1)])
            subsets = True
        elif subsets is True:
            dim -= 1

        filtration = alpha_filtration.create_alpha_shape(
            input_data, dim, weight, subsets
        ).create_filtration(no_squared)

        if check_acyclicity:
            filtration.check_for_relative_filtration()

        filtration.indexize()
        if save_to is None:
            save_to = tempfile(".idiagram")
        filtration.compute_diagram_and_save(save_to, parallels)
        return PDs(save_to, "idipha")

    @staticmethod
    def from_bitmap_levelset(array, mode, parallels=1, save_to=None):
        if mode == "sublevel":
            flip_sign = False
        elif mode == "superlevel":
            array = -array
            flip_sign = True
        bitmap = bitmap_complex.BitmapComplex(array, flip_sign)
        bitmap.indexize()
        if not save_to:
            save_to = tempfile(".idiagram")
        bitmap.compute_diagram_and_save(save_to, parallels)
        return PDs(save_to, "idipha")

    @staticmethod
    def from_bitmap_distance_function(binary_pict, signed=False,
                                      metric="manhattan", parallels=1,
                                      save_to=None):
        bitmap = binarize.distance_bitmap(binary_pict, metric, signed)
        bitmap.indexize()
        if not save_to:
            save_to = tempfile(".idiagram")
        bitmap.compute_diagram_and_save(save_to, parallels)
        return PDs(save_to)

    def save(self, dest):
        shutil.copyfile(self.path, dest)

    def dth_diagram(self, d):
        if d not in self.diagrams:
            self.diagrams[d] = PD(self.path,
                                  utils.load_diagram(self.type, self.path, d))
        return self.diagrams[d]

    __getitem__ = dth_diagram

    def invoke_gui_plotter(self, d):
        subprocess.Popen([sys.executable, "-m", "homcloud.plot_PD_gui",
                          "-d", str(d), self.path])


@forwardable()
class PD(object):
    def __init__(self, path, pd):
        self.path = path
        self.pd = pd
        self.spatial_searcher = None
        self._pairs = None

    def __repr__(self):
        return "PD(path=%s, d=%d)" % (self.path, self.pd.degree)

    def birth_death_times(self):
        return self.pd.births, self.pd.deaths

    def_delegators("pd", "births,deaths,birth_positions,death_positions,essential_births,essential_birth_positions")

    def histogram(self, x_range=None, x_bins=128, y_range=None, y_bins=None,):
        return Mesh(x_range, x_bins, y_range, y_bins, self.pd).pd_histogram(self.pd)

    def phtrees(self, save_to=None):
        if save_to is None:
            save_to = tempfile(".pht")

        if self.pd.index_map.type() != index_map.MapType.alpha:
            raise(RuntimeError("PHTrees is computable only from an alpha filtration"))

        if self.pd.degree != self.pd.index_map.dimension - 1:
            raise(RuntimeError("degree should equal dimension - 1"))

        pht = full_ph_tree.PHTrees(self.pd, self.pd.degree)
        pht.construct_tree()
        with open(save_to, "wb") as f:
            full_ph_tree.write_phtree(f, pht, self.pd)

        return PHTrees(save_to)

    def pair(self, nth):
        return Pair(
            self,
            self.pd.masked_birth_indices[nth],
            self.pd.masked_death_indices[nth],
            self.pd.birth_positions[nth] if self.pd.index_map else None,
            self.pd.death_positions[nth] if self.pd.index_map else None
        )

    def pairs(self):
        return [self.pair(n) for n in range(len(self.pd.births))]

    def nearest_pair_to(self, x, y):
        return self.get_spatial_searcher().nearest_pair(x, y)

    def get_spatial_searcher(self):
        if not self.spatial_searcher:
            self.spatial_searcher = SpatialSearcher(self.pairs(), self.births,
                                                    self.deaths)
        return self.spatial_searcher

    def pairs_in_rectangle(self, xmin, xmax, ymin, ymax):
        return self.get_spatial_searcher().in_rectangle(xmin, xmax, ymin, ymax)

    @staticmethod
    def empty():
        return PD("", homdiag.PD.empty_pd())

    def slice_histogram(self, x1, y1, x2, y2, width, bins=100):
        transl, mat = plot_PD_slice.transform_to_x_axis(
            np.array([x1, y1]), np.array([x2, y2])
        )
        xy = np.dot(mat, np.array([self.births, self.deaths]) - transl.reshape(2, 1))
        mask = (xy[0, :] >= 0) & (xy[0, :] <= 1) & (np.abs(xy[1, :]) < width / 2)
        values, edges = np.histogram(xy[0, mask], bins, (0, 1))
        return SliceHistogram(values, edges, x1, y1, x2, y2)


class SliceHistogram(object):
    def __init__(self, values, edges, x1, y1, x2, y2):
        self.values = values
        self.edges = edges
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def plot(self, left_label=None, right_label=None, logscale=False):
        ax = plt.gca()
        width = self.edges[1] - self.edges[0]
        plt.bar(self.edges[:-1], self.values, width=width, bottom=0, align="edge",
                log=logscale)
        plt.xticks([0, 1])
        plt.xlim(0, 1)
        ax.set_xticklabels([
            plot_PD_slice.construct_label(left_label, self.x1, self.y1),
            plot_PD_slice.construct_label(right_label, self.x2, self.y2),
        ])



class Pair(object):
    def __init__(self, diagram, birth_index, death_index,
                 birth_position, death_position):
        self.diagram = diagram
        self.index_map = self.diagram.pd.index_map
        self.birth_index = birth_index
        self.death_index = death_index
        self.birth_position = birth_position
        self.death_position = death_position

    def birth_time(self):
        return self.index_map.levels[self.birth_index]

    def death_time(self):
        return self.index_map.levels[self.death_index]

    birth = property(birth_time)
    death = property(death_time)

    @property
    def birth_pos(self):
        return self.birth_position

    @property
    def death_pos(self):
        return self.death_position

    def optimal_volume(self, cutoff_radius=None, solver=None,
                       constrain_birth=False, num_retry=4,
                       integer_programming=False):
        optimal_volume, status = VolumeOptimalCycleFinder(
            self.diagram.pd, self.diagram.pd.degree, cutoff_radius,
            solver, constrain_birth, num_retry, integer_programming
        ).query(
            self.index_map.simplex(self.birth_index),
            self.index_map.simplex(self.death_index)
        )
        if not optimal_volume:
            raise(RuntimeError(status.message))

        return OptimalVolume(self.diagram, optimal_volume)

    def __eq__(self, other):
        return (isinstance(other, Pair) and
                self.diagram == other.diagram and
                self.birth_index == other.birth_index and
                self.death_index == other.death_index)

    def __repr__(self):
        return "Pair({}, {})".format(self.birth_time(), self.death_time())

    def lifetime(self):
        return self.death_time() - self.birth_time()


@forwardable()
class OptimalVolume(object):
    def __init__(self, diagram, optimal_volume):
        self.diagram = diagram
        self.optimal_volume = optimal_volume

    def_delegators("optimal_volume", "points,boundary_points,boundary")

    def birth_time(self):
        return self.optimal_volume.birth.time()

    def death_time(self):
        return self.optimal_volume.death.time()

    def simplices(self):
        return self.optimal_volume.volume_simplices()

    def children(self):
        return [
            Pair(self.diagram, birth.index, death.index,
                 birth.point_coords(), death.point_coords())
            for (birth, death) in self.optimal_volume.children_pairs
        ]

    def to_dict(self):
        return self.optimal_volume.to_jsondict()

    def birth_position(self):
        return self.optimal_volume.birth.point_coords()

    def death_position(self):
        return self.optimal_volume.death.point_coords()

    birth_pos = birth_position
    death_pos = death_position


@forwardable()
class Histogram(object):
    def __init__(self, orig):
        self.orig = orig

    @property
    def values(self):
        return self.orig.values

    @values.setter
    def values(self, values):
        self.orig.values = values

    def_delegators('orig', 'x_range,y_range,x_bins,y_bins,xedges,yedges')

    def plot(self, colorbar={}, style="colorhistogram", title="",
             unit_name=None, font_size=None, aspect="auto", fig=None):
        plotter = plot_PD.PDPlotter.find_plotter(style)(
            self.orig, self._zspec(colorbar), plot_PD.AuxPlotInfo(
                title, unit_name, font_size, aspect
            )
        )
        plotter.plot(*self._fig_ax(fig))

    @staticmethod
    def _zspec(colorbar):
        type = colorbar.get("type", "linear")
        vmax = colorbar.get("max")
        vmin = colorbar.get("min")
        if "colormap" in colorbar:
            colormap = plt.get_cmap(colorbar["colormap"])
        else:
            colormap = None
        if type == "linear":
            return plot_PD.ZSpec.Linear(vmax, vmin, colormap)
        elif type == "log":
            return plot_PD.ZSpec.Log(vmax, vmin, colormap)
        elif type == "loglog":
            return plot_PD.ZSpec.LogLog(vmax, vmin, colormap)
        elif type == "linear-midpoint":
            return plot_PD.ZSpec.LinearMidPoint(colorbar["midpoint"],
                                                vmax, vmin, colormap)
        elif type == "power":
            return plot_PD.ZSpec.Power(colorbar["p"], vmax, vmin, colormap)
        else:
            raise(RuntimeError("unknown colorbar type: {}").format(type))

    @staticmethod
    def _fig_ax(fig):
        if fig is None:
            return plt.subplots()
        else:
            return (fig, fig.add_subplot())


class PHTrees(object):
    def __init__(self, path):
        self.path = path
        self.resolver = None

    def save(self, dest):
        shutil.copyfile(self.path, dest)

    def _resolver(self):
        if self.resolver is None:
            with open(self.path, "rb") as f:
                self.resolver = full_ph_tree.TreeResolver.load(f)
        return self.resolver

    def nodes_of(self, pairs):
        return [self._resolver().phtree.nodes[pair.death_index]
                for pair in pairs]

    def query_point(self, x, y):
        return self._resolver().query_node(x, y)

    def query_rectangle(self, xmin, xmax, ymin, ymax):
        return self._resolver().query_nodes_in_rectangle(xmin, xmax, ymin, ymax)

    def show_volume(self, node):
        path = tempfile(".vtk")
        node.draw_descendants_volumes(node.index_map.points,
                                      path, False, False)
        utils.invoke_paraview(path, nohup=False)

    def show_volumes(self, nodes):
        path = tempfile(".vtk")
        self._resolver().draw_volumes_of_nodes(nodes, path, True, True)
        utils.invoke_paraview(path, nohup=False)

    def show_volume_at(self, x, y):
        self.show_volume(self.query_point(x, y))

    def show_volumes_in(self, xmin, xmax, ymin, ymax):
        self.show_volumes(self.query_rectangle(xmin, xmax, ymin, ymax))


class MaskHistogram(Histogram):
    def filter_pairs(self, pairs):
        return [pair for pair in pairs
                if self.orig.value_at(pair.birth_time(), pair.death_time())]

    def filter_pd(self, pd):
        return self.filter_pairs(pd.pairs())


class Mesh(object):
    def __init__(self, x_range, xbins, y_range=None, ybins=None, pd=None):
        self.rulers = homdiag.Ruler.create_xy_rulers(x_range, xbins,
                                                     y_range, ybins, pd)
        self._histoinfo = None

    def pd_histogram(self, pd):
        return Histogram(homdiag.PDHistogram(pd, *self.rulers))

    def histogram_from_vector(self, vector):
        return Histogram(homdiag.Histogram.reconstruct_from_vector(
            vector, self.histoinfo()
        ))

    def mask_from_vector(self, vector):
        return MaskHistogram(homdiag.BinaryHistogram.reconstruct_from_vector(
            vector, self.histoinfo()
        ))

    def mask_from_2darray(self, array):
        return MaskHistogram(homdiag.BinaryHistogram(
            array, self.rulers[0].edges(), self.rulers[1].edges()
        ))

    def histoinfo(self):
        self._histoinfo = self._histoinfo or vectorize_PD.histogram_info_dict(
            homdiag.PDHistogram(homdiag.PD.empty_pd(), *self.rulers)
        )
        return self._histoinfo


class PIVectorizerMesh(Mesh):
    def __init__(self, x_range, xbins, y_range=None, ybins=None, weight=None, sigma=None):
        super(PIVectorizerMesh, self).__init__(x_range, xbins, y_range, ybins, None)
        self.weight = self.weight_function(weight)
        self.sigma = sigma

    @staticmethod
    def weight_function(params):
        if params == "none":
            return lambda b, d: 1.0
        if isinstance(params, tuple):
            if params[0] == "atan":
                return vectorize_PD.atan_weight_function(params[1], params[2])
            if params[0] == "linear":
                return vectorize_PD.linear_weight_function(1.0 / params[1])
            raise(ValueError("Unknown weight type: {}".format(params[0])))
        return params

    def vectorize(self, pd):
        histo = homdiag.PDHistogram(pd, *self.rulers)
        histo.apply_weight(self.weight)
        histo.apply_gaussian_filter(self.sigma)
        return histo.vectorize()


class BitmapPHTreesPair(object):
    def __init__(self, path):
        self._dic = None
        self.path = path

    def sign(self):
        return -1 if self.dic["sign-flipped"] else 1

    def PDs(self):
        if os.path.splitext(self.path)[1] == "json":
            raise RuntimeError("PDs cannot be created from a json file")
        return PDs(self.path)

    def dim_0_trees(self):
        return BitmapPHTrees(self, self.dic["lower"])

    def codim_1_trees(self):
        return BitmapPHTrees(self, self.dic["upper"])

    @property
    def dic(self):
        if self._dic is not None:
            return self._dic

        _, ext = os.path.splitext(self.path)
        if ext == ".json":
            with open(self.path) as f:
                self._dic = json.load(f)
        else:
            with open(self.path, "rb") as f:
                self._dic = msgpack.load(f, encoding="UTF-8")

        return self._dic

    @staticmethod
    def from_bitmap_levelset(array, mode, save_to=None):
        is_superlevel = mode == "superlevel"
        lower, upper = pict_tree.construct_mergetrees(array, is_superlevel)
        dic = pict_tree.construct_dict(array.ndim, is_superlevel, lower, upper)
        save_to = BitmapPHTreesPair._save(dic, save_to)
        return BitmapPHTreesPair(save_to)

    @staticmethod
    def from_bitmap_distance_function(array, signed=False, metric="manhattan",
                                      save_to=None):
        bitmap = binarize.distance_bitmap(array, metric, signed)
        lower, upper = pict_tree.construct_mergetrees(bitmap.array, False)
        dic = pict_tree.construct_dict(array.ndim, False, lower, upper)
        save_to = BitmapPHTreesPair._save(dic, save_to)
        return BitmapPHTreesPair(save_to)

    @staticmethod
    def _save(dic, save_to):
        if save_to is None:
            save_to = tempfile(".p2mt")
        _, ext = os.path.splitext(save_to)
        if ext == ".json":
            with open(save_to, "w") as f:
                json.dump(dic, f)
        else:
            with open(save_to, "wb") as f:
                msgpack.dump(dic, f)
        return save_to


class BitmapPHTrees(object):
    def __init__(self, treespair, dic):
        self.treespair = treespair
        self.dic = dic
        self.nodes = [self.Node(self, d) for d in self.dic["nodes"].values()]
        self.essential_nodes = [node for node in self.nodes if node.essential()]
        self.nonessential_nodes = [node for node in self.nodes if not node.essential()]
        self.spatial_searcher = self.build_spatial_searcher()

    def build_spatial_searcher(self):
        noness_nodes = [node for node in self.nodes if not node.essential()]
        births = np.array([node.birth_time() for node in noness_nodes])
        deaths = np.array([node.death_time() for node in noness_nodes])
        return SpatialSearcher(noness_nodes, births, deaths)

    def degree(self):
        return self.dic["degree"]

    def nearest_pair_node(self, x, y):
        return self.spatial_searcher.nearest_pair(x, y)

    def pair_nodes_in_rectangle(self, xmin, xmax, ymin, ymax):
        return self.spatial_searcher.in_rectangle(xmin, xmax, ymin, ymax)

    class Node(object):
        def __init__(self, mt, dic):
            self.mt = mt
            self.dic = dic

        def birth_time(self):
            return self.dic["birth-time"]

        def death_time(self):
            if not self.essential():
                return self.dic["death-time"]
            return self.mt.treespair.sign() * np.inf

        def lifetime(self):
            return self.death_time() - self.birth_time()

        def birth_position(self):
            return self.dic["birth-pixel"]

        def death_position(self):
            return self.dic["death-pixel"]

        def volume(self):
            return self.dic["volume"]

        def essential(self):
            return self.dic["death-time"] is None

        def parent(self):
            parent_id = self.dic["parent"]
            return BitmapPHTrees.Node(self.mt,
                                      self.mt.dic["nodes"][parent_id])

        def children(self):
            return [
                BitmapPHTrees.Node(self.mt, self.mt.dic["nodes"][child_id])
                for child_id in self.dic["children"]
            ]

        def __eq__(self, other):
            return ((self.mt == other.mt) and
                    (self.dic["id"] == other.dic["id"]))


def draw_pixels_on_2d_image(image, pixels, color, alpha=1.0):
    for (y, x) in pixels:
        blended_color = tuple((np.array(image.getpixel((x, y))) * (1 - alpha) +
                               np.array(color) * alpha).astype(int))
        image.putpixel((x, y), blended_color)
    return image


def draw_volumes_on_2d_image(nodes, image, color, alpha=1.0,
                             birth_position=None, death_position=None):
    image = to_pil_image(image)
    pixels = set(chain.from_iterable([map(tuple, node.volume()) for node in nodes]))
    draw_pixels_on_2d_image(image, pixels, color, alpha)
    if birth_position:
        draw_pixels_on_2d_image(image, [node.birth_position() for node in nodes],
                                birth_position)
    if death_position:
        draw_pixels_on_2d_image(image, [node.death_position() for node in nodes],
                                death_position)
    return image


def draw_birthdeath_pixels_2d(
        pairs, image, draw_birth=False, draw_death=False, draw_line=False, scale=1,
        marker_type="filled-diamond", marker_size=1, with_label=False,
        birth_color=(255, 0, 0), death_color=(0, 0, 255), line_color=(0, 255, 0),
):
    image = to_pil_image(image)

    output_image, marker_drawer = view_index_pict.setup_images(
        image, scale, not with_label
    )
    marker_drawer.scale = scale
    marker_drawer.marker_size = marker_size
    marker_drawer.should_draw_birth_pixel = draw_birth
    marker_drawer.should_draw_death_pixel = draw_death
    marker_drawer.should_draw_line = draw_line
    marker_drawer.no_label = not with_label
    marker_drawer.birth_color = birth_color
    marker_drawer.death_color = death_color
    marker_drawer.line_color = line_color

    for pair in pairs:
        marker_drawer.draw_pair(pair)

    return output_image


def to_pil_image(image):
    if isinstance(image, str):
        return view_index_pict.open_image(image)
    elif isinstance(image, np.ndarray):
        upper = np.max(image)
        lower = np.min(image)
        return Image.fromarray(((image - lower) / (upper - lower) * 250).astype(np.uint8))
    elif isinstance(image, Image.Image):
        return image
    else:
        raise(RuntimeError("{} cannot be converted to PIL.Image".format(repr(image))))
